/*
 * Copyright (C) 1984-2021  Mark Nudelman
 *
 * You may distribute under the terms of either the GNU General Public
 * License or the Less License, as specified in the README file.
 *
 * For more information, see the README file.
 */


/*
 * Include file for interfacing to position.c modules.
 */
#define TOP             (0)
#define TOP_PLUS_ONE    (1)
#define BOTTOM          (-1)
#define BOTTOM_PLUS_ONE (-2)
#define MIDDLE          (-3)
